static char rcsid[] = "@(#)$Id: builtin++.c,v 2.28 2014/05/03 17:33:38 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.28 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 *
 *  Initially written by: Michael Elkins <elkins@aero.org>, 1995
 *****************************************************************************/

#include "def_elm.h"
#include "pager.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ui");

static int search_it P_((struct stringbuffer * bout, int start_idx,
			 struct string * search_pattern));

static int search_it(bout,start_idx,search_pattern)
     struct stringbuffer * bout; 
     int start_idx;
     struct string * search_pattern;
{
    int pg_flags;
    int ret = 0;
    int idx = start_idx;

    while (idx < linecount_stringbuffer(bout) && !ret) {
	struct pager_lineext *lineext = NULL;
	struct string * buffer = 
	    get_line_from_stringbuffer(bout,idx,&pg_flags,NULL,NULL,
				       &lineext);

	if (find_pattern_from_string(buffer,search_pattern,1))
	    ret = idx;

	while (lineext && !ret) {
	    struct string * buffer2 = 
		get_lineext_from_stringbuffer(&lineext,&pg_flags,NULL,NULL);

	    if (find_pattern_from_string(buffer2,search_pattern,1))
		ret = idx;

	    free_string(&buffer2);
	}
	
	idx++;
	free_string(&buffer);
	if (lineext)
	    free_pager_lineext(&lineext);
    }
    DPRINT(Debug,3,
	   (&Debug,"Builtin pager: Search from line %d %s (%d)\n",
	    start_idx, 
	    ret ? "SUCCEED" : "FAILED",
	    ret));
    return ret;
}

int builtinplusplus (bout,pager_page, is_helper, pager_cmds)
     struct stringbuffer * bout;
     struct pager_page * pager_page;
     int is_helper;
     struct elm_commands *pager_cmds;
{   
    int                    idx = 0;
    struct pager_lineext * lineext = NULL;

    int is_end = 0;

    /* Current line for outputting and position on it ... */
    struct string * buffer = NULL;
    int                ret = '\0';
    int             pg_flags = 0;
    int X = 0;
    struct string * search_pattern = NULL;        /* Last seach string */
    struct pager_range  * pager_range = NULL;

    struct pg_wrap_info * info = new_pg_wrap_info();
    int virtual_line           = 0;  /* Index for next pg_wrap_info */

    int     force_word_wrap = 0;
    int     p_width = pager_paragraph_width;

    int     goto_idx = 0, goto_X = 0;
    struct pager_lineext * goto_lineext = NULL;
    struct pager_lineext *lineext_top = NULL;  /* line extension on top of display */


    clear_error();
    
    for (;;) {
	int virtual_line_top;	    
	int idx_top;             /* line number of top of display  */

	int ch;                  /* command key read */
	int search_matches = 0;  /* First idx on page where search matches */
	int lines;               /* Number of visible lines printed
				    so far */
	int FF_seen;            /* If page is breaken because of FF */
	int LINES, COLUMNS;
	int scroll_line = 0;

	int start_of_line = 1;
	int chars = 0; /* Number of characters on current line */
	int mayclear = 1;
	int word_wrapped;
	int preformatted;
	int joined_line;
	int p_start_COL;
	int buffer_len = -1;

	if (goto_idx >= 0 && goto_X >= 0) {
	goto_page:

	    DPRINT(Debug,5,(&Debug, 
			    "builtin++ -- Going page idx = %d, X=%d\n",
			    goto_idx,goto_X));
	    if (buffer)
		free_string (&buffer);
	    idx = goto_idx;
	    X   = goto_X;
	    
	    if (lineext)
		free_pager_lineext(&lineext);

	    if (goto_lineext) {
		lineext = goto_lineext; goto_lineext = NULL;

		DPRINT(Debug,5,(&Debug, 
				"builtin++ -- X=%d, going to line extension on buffer line %d\n",
				X,idx));

		buffer = get_lineext_from_stringbuffer(&lineext,&pg_flags,
						       NULL,
						       &pager_range);
		
		buffer_len = string_len(buffer);

	    } else {
		if (idx < linecount_stringbuffer(bout)) {

		    DPRINT(Debug,5,(&Debug, 
				    "builtin++ -- X=%d, going to buffer line %d\n",
				    X,idx));

		    buffer = get_line_from_stringbuffer(bout,idx,&pg_flags,
							NULL,
							&pager_range,
							&lineext); 
		    buffer_len = string_len(buffer);
		} else {
		    DPRINT(Debug,5,(&Debug, 
				    "builtin++ -- No buffer line %d\n",idx));
		    buffer_len = -1;
		}
		
	    }

	    if (X > buffer_len) {
		DPRINT(Debug,5,(&Debug, 
				"builtin++ -- No X=%d (len=%d) on buffer line %d%s\n",
				X,buffer_len,idx,
				lineext ? ", have line extension" : ""
				));
	    } else if (X == buffer_len) {
		DPRINT(Debug,5,(&Debug, 
				"builtin++ -- X=%d at end of line %d%s\n",
				X,idx,
				lineext ? ", have line extension" : ""
				));
	    }
		    	
	    goto_idx = -1;
	    goto_X   = -1;
	    menu_ClearScreen(pager_page->root);
	}

    new_page:
    resize_mark:

	/* Check these also when page is empty !!! */
	if (menu_resized(pager_page->root)) {

	    menu_get_sizes(pager_page->root, &LINES, &COLUMNS);

	    DPRINT(Debug,5,(&Debug, 
			    "builtin++ top -- root resized\n"));

	}

	if (menu_need_redraw(pager_page->root)) {
	    DPRINT(Debug,5,(&Debug, 
			    "builtin++ top -- root need redraw\n"));

	}

	idx_top = idx;       /* line number of top of display  */
	if (lineext_top)
	    free_pager_lineext(&lineext_top);
	lineext_top = lineext;  /* line extension on top of display */
	if (lineext_top)
	    inc_pager_lineext_refcount(lineext_top);

	virtual_line_top = virtual_line;
	lines = 0;           /* Number of visible lines printed
				so far */
        FF_seen = 0;
	word_wrapped = 0;
	preformatted = 0;
	joined_line = 0;
	p_start_COL = 0;
	chars       = 0;    /* On beginning of line */

	menu_ClearScreen(pager_page->root);

    cr_restart:                  /* so that we can handle CR specially --
				    ie. don't reset page variables 
				 */

	menu_get_sizes(pager_page->root, &LINES, &COLUMNS);

	


 
	DPRINT(Debug,5,(&Debug, 
			"builtin++ top  idx = %d, scroll_line =%d, lines = %d\n",
			idx, scroll_line, lines));

	while ((idx < linecount_stringbuffer(bout)) &&
	       (lines < LINES-1 || 
		(scroll_line && lines < LINES)
		) && !FF_seen) {
	    
	    int this_flags = 0;
	    enum span_result span_line_result;
	    int was_start_of_line = start_of_line;
	    int range_flags = 0;

	    if (menu_resized(pager_page->root)) {
		menu_get_sizes(pager_page->root, &LINES, &COLUMNS);

		if (buffer)
		    free_string(&buffer);
	    
		idx= idx_top;
		goto new_page;
	    }

	    if (menu_need_redraw(pager_page->root)) {

		if (buffer)
		    free_string(&buffer);

		idx= idx_top;
		goto new_page;

	    }
    
	    DPRINT(Debug,9,(&Debug, 
			    "builtin++ before line  idx = %d (virtual %d) X=%d start_of_line=%d\n",
			    idx,virtual_line,X,start_of_line));

	    if (!buffer || 
		X >= (buffer_len = string_len(buffer))) {
		int range_changed = 0;
		/* caller of get_line_from_stringbuffer is expected to 
		   free_string () resulting buffer */

		if (buffer) {
		    if (X >= buffer_len && !lineext)
			idx++;
		    if (idx >= linecount_stringbuffer(bout))
			break;
		    free_string (&buffer);
		}
		X = 0;

		/* There is no pending data to output, so we need to grab 
		 * another line.
		 */

		if (lineext) {
		    buffer = get_lineext_from_stringbuffer(&lineext,&pg_flags,
							   &range_changed,
							   &pager_range);

		    buffer_len = string_len(buffer);

		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ new line extension on line %d -- len = %d\n",
				    idx,buffer_len));
							   
		} else {

		    buffer = get_line_from_stringbuffer(bout,idx,&pg_flags,
							&range_changed,
							&pager_range,
							&lineext);   
		    buffer_len = string_len(buffer);
		    
		    
		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ new line idx=%d -- len = %d\n",
				    idx,buffer_len));

		}

		if (!start_of_line && range_changed) {  
		    /* Need new line */

		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ wrapping line because of pager range change\n"));


		    start_of_line = 1;
		    goto wrap_line;
		}
	    }

	    if (pager_range) {
		range_flags = get_pager_range_flags(pager_range);

		DPRINT(Debug,9,(&Debug, 
				"builtin++ pager range: quote_level=%d range_flags=%d\n",
				get_pager_range_quote_level(pager_range),range_flags));
	    }

	    if (start_of_line) {
		DPRINT(Debug,10,(&Debug, 
				"builtin++ pager range: virtual_line=%d, start of line\n",
				 virtual_line));

		record_pg_wrap_info(info,&virtual_line,idx,X,lineext);
	    }
	   
	    if (search_matches &&
		search_matches == idx)
		this_flags = pg_STANDOUT;
	    else
		this_flags = pg_flags;
	    
	    if (start_of_line && pager_range) {
		uint16 ch = 0;   /* First character may be needed */

		if (buffer_len > 0)
		    ch = give_unicode_from_string(buffer,0);

		span_line_result = span_range_prefix(pager_page->root,
						     lines,&chars,ch,
						     pager_range,&this_flags,
						     &mayclear);
					
		DPRINT(Debug,9,(&Debug, 
				"builtin++ span_range_prefix result = %d, ",
				span_line_result));

		start_of_line = 0;
		p_start_COL = chars;

		if (X >= buffer_len) {
		    DPRINT(Debug,9,(&Debug, 
				    "no buffer left (X=%d buffer_len=%d), ",
				    X,buffer_len));
		    goto buffer_handled;
		}

	    } else {
		if (start_of_line) {
		    DPRINT(Debug,10,(&Debug, 
				     "builtin++ | No pager range on beginning of line\n"));
		    
		    start_of_line = 0;
		    p_start_COL = 0;
		}

		if (joined_line) {
		    DPRINT(Debug,10,(&Debug, 
				    "builtin++ | this is joined line\n"));
		}

		if (0 != (range_flags & PR_WORD_WRAP) 
		    || 
		    (force_word_wrap && 0 == (range_flags & PR_PREFORMAT))
		    ) {
	       
		    span_line_result = span_words(pager_page->root,lines,&chars,
						  buffer,&X,&mayclear,
						  this_flags,
						  COLUMNS    /* max_width*/,
						  buffer_len /* max_len */,
						  joined_line,
						  p_start_COL,
						  0 == (range_flags & 
							PR_MAX_WIDTH) ? 
						  p_width : 0 );

		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ span_words result = %d, ",
				    span_line_result));


		} else {
		    
		    /* This is the part of the code that actually displays on 
		     * the screen 
		     */
		    
		    span_line_result = span_line(pager_page->root,lines,&chars,
						 buffer,&X,&mayclear,this_flags,
						 COLUMNS    /* max_width*/,
						 buffer_len /* max_len */);
		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ span_line result = %d, ",
				    span_line_result));
		    
		}
	
	    buffer_handled:
		if (lineext) {
		    DPRINT(Debug,9,(&Debug," line extension (joining), "));

		    joined_line = 1;
		} else if (0 == (range_flags & PR_JOIN_LINES)) {
		    DPRINT(Debug,9,(&Debug," line wrapping, "));
		    start_of_line = 1;  /* Need new line */
		} else {
		    DPRINT(Debug,9,(&Debug," joining lines, "));
		    joined_line = 1;
		}
	    }

	    switch (span_line_result) {
	    case span_failure:
		
		if (!was_start_of_line) {
		    DPRINT(Debug,9,(&Debug, 
				    "failure, trying restart from next line\n"));
		    
		    
		    start_of_line = 1;
		} else {

		default:
		    DPRINT(Debug,9,(&Debug, 
				    "failure, OOPS\n"));
		    
		    goto read_key;
		}
		break;
				    
	    case span_ok:
		DPRINT(Debug,9,(&Debug, 
				"ok\n"));
		break;
	    case span_FF_seen:
		DPRINT(Debug,9,(&Debug, 
				"FF_seen\n"));
		FF_seen = 1;
		start_of_line = 1;
		break;
	    case span_EOS:
		DPRINT(Debug,9,(&Debug, 
				"at end of string\n"));
		break;
	    case span_NL_seen:
		DPRINT(Debug,9,(&Debug, 
				"new line seen\n"));
		start_of_line = 1;

		break;
	    case span_wrapping:
		DPRINT(Debug,9,(&Debug, 
				"wrapping\n"));
		start_of_line = 1;

		if (0 != (range_flags & PR_PREFORMAT))
		    preformatted = 1;
		break;
	    case span_wordwrapped:
		DPRINT(Debug,9,(&Debug, 
				"word wrapped\n"));
		start_of_line = 1;
		word_wrapped = 1;
		break;
	    }
	    
	    /* Make sure that carbage from previous
	       output is not left */
	    if (mayclear) {
		menu_CleartoEOLN(pager_page->root);
		mayclear = 0;
	    }
	    
	    if (start_of_line) {
	    wrap_line:
		menu_Writechar(pager_page->root,'\r');		    
		menu_Writechar(pager_page->root,'\n');		    
		lines++;
		chars = 0;

		joined_line = 0;
		p_start_COL = 0;
	    }

	    if (scroll_line) {
		int x, y;
		
		menu_GetXYLocation (pager_page->root,&x, &y);
		if (x != lines) {
		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ line scrolled, lines=%d (not %d)\n",
				    x,lines));
		    lines = x;
		    scroll_line--;
		}
	    }
	}

	if (!start_of_line) {
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    
	    lines++;
	}

	menu_CleartoEOS(pager_page->root);
	if (lines < LINES-1 && FF_seen) {
	    menu_StartXX(pager_page->root,pg_BOLD);
	    menu_PutLineX (pager_page->root,
			   lines, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmBuiltinNewPage,
				   "~ (new page) "));
	    menu_EndXX(pager_page->root,pg_BOLD);  
	    menu_CleartoEOLN(pager_page->root);
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    

	    lines++;
	}
	while (lines < LINES-1) {
	    menu_Writechar(pager_page->root,'~');
	    menu_CleartoEOLN(pager_page->root); 
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    
	    
	    lines++;
	}

    reprompt:
	menu_StartXX(pager_page->root,pg_STANDOUT);
	
	if (idx >= linecount_stringbuffer(bout)) {
	    if (is_helper)
		menu_PutLineX (pager_page->root,LINES-1, 0, 
			       CATGETS(elm_msg_cat, ElmSet, ElmBuiltinHelpEnd,
				       "Press 'q' or <return> to return:"));

	    else
		menu_PutLineX (pager_page->root,LINES-1, 0, 
			       CATGETS(elm_msg_cat, ElmSet, ElmBuiltinCommandi,
				       "Command ('i' to return to index):"));
	    is_end = 1;

	} else {

	    int relative;
	    enum user_level_v ul = 
		give_dt_enumerate_as_int(&user_level);

	    if (linecount_stringbuffer(bout) < 1)
		relative = 100;
	    else
		relative = 100 * idx / linecount_stringbuffer(bout);
	    
	    if (force_word_wrap && word_wrapped) {
		menu_PutLineX(pager_page->root,LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInWordWrapped,
				      "Word wrapped (line %d-%d/%d, you've seen %d%%). Press 'W' to disable:"),
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);	      
	    } else if (force_word_wrap && preformatted) {
		menu_PutLineX(pager_page->root,LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInPreformatted,
				      "Preformatted (line %d-%d/%d, you've seen %d%%). Press 'W' to mute:"),
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);	      		
	    } else if (0 == p_width && pager_paragraph_width && word_wrapped) {
		menu_PutLineX(pager_page->root,LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInFullWidth,
				      "Full width (line %d-%d/%d, you've seen %d%%). Press 'w' to mute:"),
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);
	    } else if (is_helper) {
		menu_PutLineX(pager_page->root,
			      LINES-1, 0,CATGETS(elm_msg_cat, ElmSet, ElmBuiltHelpMore,
						 "Press <space> to continue, 'q' to return:"));
	    } else if (ul == user_level_beginner)
		menu_PutLineX(pager_page->root,LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInMore0,
				      "MORE (line %d-%d/%d, you've seen %d%%). Press <space> for more:"), 
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);
	    else
		menu_PutLineX(pager_page->root,
			      LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInMore,
				      "MORE (line %d-%d/%d, you've seen %d%%):"), 
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);
	    is_end = 0;
	}
	menu_EndXX(pager_page->root,pg_STANDOUT);
	menu_CleartoEOS(pager_page->root);

    read_key:
	ch = menu_ReadCh(pager_page->root, REDRAW_MARK|READCH_CURSOR|READCH_resize|
			 READCH_sig_char);

	if (ch != '/') {
	    if (search_pattern) {
		free_string(&search_pattern);
	    }
	}

	switch (ch) {
	case TERMCH_interrupt_char:
	    ret = 0;
	    goto finish;

	case RESIZE_MARK:
	    DPRINT(Debug,4, (&Debug, "    ... resizing\n"));
	    idx          = idx_top;
	    if (lineext)
		free_pager_lineext(&lineext);
	    lineext      = lineext_top;
	    if (lineext)
		inc_pager_lineext_refcount(lineext);
	    virtual_line = virtual_line_top;

	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);

	    goto resize_mark;

	case ' ':
	case '+':
	case PAGEDOWN_MARK:
	    if (is_end) {
		ret = ' ';
		goto finish;
	    }
	break;
	case HOME_MARK:
	case '^':      

	    idx = 0;
	if (lineext)
	    free_pager_lineext(&lineext);

	virtual_line = 0;
	    /* Avoid incrementing of idx: */
	    if (buffer)    
		free_string (&buffer);
	    break;

	case 'G': {    /* Go to last page --
	                  Not completely correct if multiline lines -
			  or folded paragraphs / lines
			*/

	    int old_idx = idx;
	    int tmp UNUSED_VAROK = virtual_line;
   
	    if (lineext)
		free_pager_lineext(&lineext);

	    idx = linecount_stringbuffer(bout)-LINES+1;
	    if (idx < 0)
		idx = 0;

	    virtual_line +=  (idx - old_idx);
	    if (virtual_line < 0)
		virtual_line = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- %04x virtual_line=%d > %d\n",
			     ch,tmp,virtual_line));

	    if (goto_pg_wrap_info(info,virtual_line,&goto_idx,&goto_X,
				  &goto_lineext)) {
		DPRINT(Debug,10,(&Debug, 
				 "builtin++  -- used wrap info, index=%d, X=%d\n",
				 goto_idx,goto_X));

		goto goto_page;
	    } 

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- calculated index=%d, X=%d\n",
			     idx,X));
	   
	    lines = 0;
	    /* Avoid incrementing of idx: */
	    if (buffer)    
		free_string (&buffer);
	    menu_ClearScreen(pager_page->root);
	}
	    break;
	case FIND_MARK:
	case '/':
	    menu_ClearLine (pager_page->root,LINES-1);
	    {
		int code;
		
		code = optionally_enter2(pager_page->root,
					 &search_pattern,
					 LINES-1, 0, 
					 OE_REDRAW_MARK|
					 OE_SIG_CHAR /* Ctrl-C */, 
					 CATGETS(elm_msg_cat, ElmSet, 
						 ElmBuiltinSearch,
						 "Search: "));

		if (REDRAW_MARK == code)
		    goto redraw;
		if (-1 == code) /* Ctrl-C */
		    goto reprompt;
		if (code != 0)
		    goto quit;
		if (search_pattern && string_len(search_pattern)) {
		    int start_idx;
		    /* Continue search from previous match */
		    if (!search_matches)
			start_idx = idx_top;
		    else if (search_matches < linecount_stringbuffer(bout))
			start_idx = search_matches+1;
		    else
			start_idx = idx_top;
		    search_matches = search_it(bout,start_idx,
					       search_pattern);
		    if (search_matches) {
			int old_idx = idx;
			int tmp UNUSED_VAROK = virtual_line;

			if (search_matches > idx_top && 
			    search_matches < idx_top + LINES -4) {
			    idx = idx_top;
			    virtual_line = virtual_line_top;
			} else {
			    idx = search_matches - LINES / 2;

			    virtual_line +=  (idx - old_idx);
			}
			/* virtual_line not calculated correctly */

			DPRINT(Debug,10,(&Debug, 
					 "builtin++  -- search, virtual_line=%d > %d\n",
					 tmp,virtual_line));


			if (lineext)
			    free_pager_lineext(&lineext);

			if (idx < 0)
			    idx = 0;
			if (virtual_line < 0)
			    virtual_line = 0;
			
			/* Avoid incrementing of idx: */
			if (buffer)    
			    free_string (&buffer);
			goto new_page; /* and draw match */
		    } else {
			int relative;
			enum user_level_v ul = 
			    give_dt_enumerate_as_int(&user_level);

			if (linecount_stringbuffer(bout) < 1)
			    relative = 100;
			else
			    relative = 100 * idx / linecount_stringbuffer(bout);
			
			menu_ClearLine (pager_page->root,LINES-1);
			menu_StartXX(pager_page->root,pg_STANDOUT);
			if (ul == user_level_beginner)
			    menu_PutLineX (pager_page->root,LINES-1, 0, 
					   CATGETS(elm_msg_cat, ElmSet, ElmBuiltInNotFound0,
						   "NOT FOUND! (line %d-%d/%d, you've seen %d%%). Press <space> for more:"), 
				      idx_top+1,idx,linecount_stringbuffer(bout),
				      relative);
		    else
			menu_PutLineX (pager_page->root,LINES-1, 0, 
				       CATGETS(elm_msg_cat, ElmSet, 
					       ElmBuiltInNotFound,
					       "NOT FOUND! (line %d-%d/%d, you've seen %d%%):"), 
				       idx_top+1,idx,linecount_stringbuffer(bout),
				  relative);		    
			menu_EndXX(pager_page->root,pg_STANDOUT);
			goto read_key;
		    }
		} 
		goto reprompt;
	    }	    

	case PAGEUP_MARK:
	case '-': {
	    int tmp UNUSED_VAROK = virtual_line;

	    virtual_line = virtual_line_top - LINES+1;
	    if (virtual_line < 0)
		virtual_line = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- %04x virtual_line=%d > %d\n",
			     ch,tmp,virtual_line));

	    if (goto_pg_wrap_info(info,virtual_line,&goto_idx,&goto_X,
				  &goto_lineext)) {
		DPRINT(Debug,10,(&Debug, 
				 "builtin++  -- used wrap info, index=%d, X=%d\n",
				 goto_idx,goto_X));

		goto goto_page;	       
	    } 

	    if (lineext)
		free_pager_lineext(&lineext);

	    idx = idx_top - LINES+1;
	    if (idx < 0)
		idx = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- calculated index=%d, X=%d\n",
			     idx,X));
	    
	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);
	}
	    break;
	case ctrl('P'): {
	    int tmp UNUSED_VAROK = virtual_line;

	    virtual_line = virtual_line_top - 1;
	    if (virtual_line < 0)
		virtual_line = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- %04x virtual_line=%d > %d\n",
			     ch,tmp,virtual_line));
	    
	    if (goto_pg_wrap_info(info,virtual_line,&goto_idx,&goto_X,
				  &goto_lineext)) {
		DPRINT(Debug,10,(&Debug, 
				 "builtin++  -- used wrap info, index=%d, X=%d\n",
				 goto_idx,goto_X));
		
		goto goto_page;	       
	    } 

	    if (lineext)
		free_pager_lineext(&lineext);

	    idx = idx_top -1;
	    if (idx < 0)
		idx = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- calculated index=%d, X=%d\n",
			     idx,X));

	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);
	}
	    break;
	case ctrl('D'):
	    if (is_end) 
		goto reprompt; /* Can't go down */

	    menu_Writechar(pager_page->root,'\r');		    
	    menu_CleartoEOLN(pager_page->root);	    

	    scroll_line = LINES / 2;

	    idx_top += LINES / 2;
	    virtual_line_top += LINES / 2;

	    goto cr_restart; /* Hanlde Ctrl-D specially */
	    	   	    
	case ctrl('U'): {
	    int tmp UNUSED_VAROK = virtual_line;

	    virtual_line = virtual_line_top - LINES / 2;
	    if (virtual_line < 0)
		virtual_line = 0;
	    
	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- %04x virtual_line=%d > %d\n",
			     ch,tmp,virtual_line));
	    
	    if (goto_pg_wrap_info(info,virtual_line,&goto_idx,&goto_X,
				  &goto_lineext)) {
		DPRINT(Debug,10,(&Debug, 
				 "builtin++  -- used wrap info, index=%d, X=%d\n",
				 goto_idx,goto_X));
		
		goto goto_page;	       
	    } 
	    
	    if (lineext)
		free_pager_lineext(&lineext);

	    idx = idx_top - LINES / 2;
	    if (idx < 0)
		idx = 0;

	    DPRINT(Debug,10,(&Debug, 
			     "builtin++  -- calculated index=%d, X=%d\n",
			     idx,X));

	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);
	}
	    break;
	case 'W':	    
	    force_word_wrap = !force_word_wrap;

	    DPRINT(Debug,4,(&Debug, " ... %s word wrapping\n",
			    force_word_wrap ? "enabled" : "disabled"));
	    
	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);

	    if (lineext)
		free_pager_lineext(&lineext);
	    lineext      = lineext_top;
	    if (lineext)
		inc_pager_lineext_refcount(lineext);

	    idx          = idx_top;
	    virtual_line = virtual_line_top;
	    goto resize_mark;

	case 'w':
	    if (p_width) 
		p_width  = 0;
	    else
		p_width  = pager_paragraph_width;
	    
	    DPRINT(Debug,4,(&Debug, " ... %s full width paragaphs\n",
			    p_width ? "disabled" : "enabled"));
	    
	    /* Avoid incrementing of idx: */
	    if (buffer)
		free_string(&buffer);

	    if (lineext)
		free_pager_lineext(&lineext);
	    lineext      = lineext_top;
	    if (lineext)
		inc_pager_lineext_refcount(lineext);

	    idx          = idx_top;
	    virtual_line = virtual_line_top;
	    goto resize_mark;


	case HELP_MARK:
	case '?': {

	    if (is_helper < 1) {
		struct elm_commands *cmds = 
		    give_builtinpager_commands();

		int ch;

	        if (pager_cmds) {
		    struct elm_commands *all_cmds =
			give_merged_commands(cmds,pager_cmds,
					     pg_BOLD,0);
		    
		    struct stringbuffer * b =
			give_commands_help(all_cmds);

		    ch = builtinplusplus(b,pager_page,is_helper+1,NULL);

		    free_stringbuffer(&b);
		    free_commands(&all_cmds);
		} else {
		    struct stringbuffer * b =
			give_commands_help(cmds);

		    ch = builtinplusplus(b,pager_page,is_helper+1,NULL);

		    free_stringbuffer(&b);
		}
		free_commands(&cmds);

		if (EOF == ch) {
		    ret = EOF;
		    goto finish;
		}
	    }	   
	}
	    /* FALLTHRU */
	redraw:
	case REDRAW_MARK:
	case ctrl('L'): 

	    if (lineext)
		free_pager_lineext(&lineext);
	    lineext      = lineext_top;
	    if (lineext)
		inc_pager_lineext_refcount(lineext);

	    idx          = idx_top;
	    virtual_line = virtual_line_top;
	    if (buffer)
		free_string(&buffer);
	    break;
	case '\n':
	case '\r':
	    if (FF_seen)
		break;   /* New page */
	
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_CleartoEOLN(pager_page->root);	    

	    if (is_helper && is_end) {
		ret = 0;
		goto finish;
	    }

	    scroll_line = 1;
	    idx_top++;
	    virtual_line++;
	    goto cr_restart; /* Hanlde CR specially */

	quit:
	case EOF:
	default:
	    ret =  (ch == 'q' || ch == 'x' ? 0 : ch);
	    goto finish;
	}
    }

 finish:

    if (lineext_top)
	free_pager_lineext(&lineext_top);
    if (lineext)
	free_pager_lineext(&lineext);
    if (goto_lineext)
	free_pager_lineext(&goto_lineext);

    if (pager_range)
	free_pager_range(&pager_range);
    if (buffer)
	free_string (&buffer);
    if (info)
	free_pg_wrap_info(&info);

    DPRINT(Debug,9,(&Debug, 
		    "builtin++ returning %d",ret));
    if (isascii(ret) && isprint(ret)) {
	DPRINT(Debug,9,(&Debug, " (%c)",ret));
    }
    DPRINT(Debug,9,(&Debug, "\n"));
    return ret;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
