static char rcsid[] = "@(#)$Id: span_range.c,v 1.5 2013/10/13 07:18:54 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.5 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "def_pager.h"

DEBUG_VAR(Debug,__FILE__,"pager");

enum span_result span_range_prefix(ctx,cur_line,cur_col,
				   first_char, pager_range, this_flags,
				   mayclear)
     struct menu_context *ctx;
     int cur_line; 
     int *cur_col;
     uint16 first_char;
     struct pager_range  * pager_range;
     int *this_flags;
     int *mayclear;     
{
    enum span_result res = span_ok;

    int CUR_col    = *cur_col;
    int THIS_flags = *this_flags;
    int MAY_clear = *mayclear;

    int START_col UNUSED_VAROK = CUR_col;

    int range_flags = get_pager_range_flags(pager_range);
    int quote_level = get_pager_range_quote_level(pager_range);  /* Original flowed quote */

    int LINES, COLUMNS;

    menu_get_sizes(ctx, &LINES, &COLUMNS);

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix: line = %d, col = %d, first_char=%04x, LINES=%d, COLUMNS=%d\n",
		     cur_line,CUR_col,first_char, LINES, COLUMNS));

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix: range_flags = %d, quote_level = %d\n",
		     range_flags,quote_level));

    if (0 != (range_flags & PR_REPLY_QUOTE))                     /* Should not happen */
	quote_level++;

    if (quote_level && quote_prefix) {
	int lastX = -1, i;
	uint16 repchar = 0x0;
	int POS = 0;

	int len = string_len(quote_prefix);

	DPRINT(Debug,15,(&Debug,
			 "span_range_prefix: (len=%d) prefix=%S\n",
			 len,quote_prefix));

	for (i = 0; i < len; i++) {
	    uint16 ch  = give_unicode_from_string(quote_prefix,i);

	    if (0 != unicode_ch(ch,UOP_noctrl) &&
		0 == unicode_ch(ch,UOP_space)) {
		repchar = ch;
		lastX = i;
	    }
	}
	
	if (len > 0)
	    first_char = give_unicode_from_string(quote_prefix,0);

	if (0 != (range_flags & PR_QUOTE_L)) {                  /* Should not happen */
	    int need_quote = 0x005B /* [ */ == first_char;
	    
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: need_quote=%d first_char=%04x\n",
			     need_quote,first_char));

	    menu_StartXX(ctx,pg_BOLD);
	    if (need_quote)
		menu_Writechar(ctx,'[');
	    menu_EndXX(ctx,pg_BOLD);

	}

	if (lastX >= 0) {
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: lastX=%d, repchar=%04x\n",
			     lastX,repchar));

	    res = span_line(ctx,cur_line,&CUR_col,quote_prefix,&POS,&MAY_clear,
			    THIS_flags,COLUMNS/3,lastX+1);

	    if (res != span_ok || POS <= lastX) {
		DPRINT(Debug,12,(&Debug,
				 "span_range_prefix: failure, pos=%d, col=%d, res=%d\n",
				 POS,CUR_col,res));
		goto failure;
	    }
	} else {
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: No printable characters on prefix\n"));
	}

	if (quote_level > 1 && repchar) {
	    int x;

	    if (THIS_flags)
		menu_StartXX(ctx,THIS_flags);


	    for (x = 0; x < quote_level - 1 && CUR_col < COLUMNS/2; x++) {
	
		menu_WriteUnicode(ctx,repchar);
		CUR_col++;       
	    }
	
	    if (THIS_flags)
		menu_EndXX(ctx,THIS_flags);

	}


	if (POS < len) {
	    res = span_line(ctx,cur_line,&CUR_col,quote_prefix,&POS,&MAY_clear,
			    THIS_flags,COLUMNS/3,len-lastX);
	    
	    if (res != span_ok || POS < len) {
		DPRINT(Debug,12,(&Debug,
				 "span_range_prefix: failure, pos=%d, col=%d, res=%d\n",
				 POS,CUR_col,res));
		goto failure;
	    }
	    
	}

    failure:
	;

    } else if (0 != (range_flags & PR_QUOTE_L)) {            /* Should not happen */
	int need_quote = 0x005B /* [ */ == first_char;
	
	DPRINT(Debug,12,(&Debug,
			 "span_range_prefix: need_quote=%d first_char=%04x\n",
			 need_quote,first_char));
	
	menu_StartXX(ctx,pg_BOLD);
	if (need_quote)
	    menu_Writechar(ctx,'[');
	menu_EndXX(ctx,pg_BOLD);
	
    }
       
    *cur_col    = CUR_col;
    *this_flags = THIS_flags;
    *mayclear   = MAY_clear;

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix=%d (col = %d -> %d, line = %d)\n",
		     res,START_col, CUR_col, cur_line));

    return res;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
